;//**   File: gdoubletostr.asm
;//**
;//**   Created by GODlike (www.ancientritual.tk)

.686
.model flat, stdcall
option casemap: none


gUIntToStr proto :dword,:dword   ;using the gIntToStr to build the string
gIntToStr  proto :dword,:dword   ;                  >>


.data

pcErrorStr db "+0.0000000e+000", 00h
sizepcErrorStr equ 16

f10pwr7 dd 10000000.0            ;10^7 = 10000000.0


.code

;==========================================================
;   int stdcall gDoubleToStr (char* pcDest, double dbl)
;==========================================================
gDoubleToStr proc pcDest:dword, dbl:qword
	;/--------------------------------------------------------------------------------------------------------\
	;| Converts a double to string.                                                                           |
	;| Output String Format: 1B sign, 9B normalized mantissa, 1B 'e', 1B exponent sign, 3B exponent (=16byte) |
	;| Warning: Make sure that the dest points to a 16 byte string                                            |
	;\--------------------------------------------------------------------------------------------------------/
	local exponent:dword            ;the dbl's exponent
	local decimal:dword             ;the dbl's decimal
	local oldCW:word                ;old fpu control word
	local newCW:word                ;new fpu control word
	push edi                        ;the pointer to the pcDest
	push esi                        ;for the rep
	push ecx                        ;for the rep
	
	mov edi, dword ptr [pcDest]     ;the edi is the dest pointer
	
	;----------------------------------------------
	;check if dbl is 0.0 and set the string manualy
	;----------------------------------------------
	fld qword ptr [dbl]
	ftst
	fnstsw ax                       ;store the fpu status word
	fwait
	fstp st(0)                      ;pop reg stack
	sahf                            ;set the flagReg
	jnz @F
		mov esi, offset pcErrorStr 
		mov ecx, sizepcErrorStr
		cld
		rep movsb [edi], [esi]        ;copy strings
		mov eax, sizepcErrorStr-1     ;the return value
		jmp return
	@@:
	
	;---------------------------------------------------------
	;check if dbl is negative set al=='+' or '-' and fabs(dbl)
	;---------------------------------------------------------
	mov al, byte ptr [dbl+7]        ;move to al the first byte of dbl
	mov cl, al                 
	and cl, 01111111b               ;make the less significant bit of the first byte...
	mov byte ptr [dbl+7], cl        ;...of dbl equal to zero (make dbl possitive)
	
	and al, 10000000b          
	test al, al                     ;if dbl is possitive set al='+' else al='-' (without brunch)
	seta al
	sub al, 01h
	and al, '+' - '-'
	add al, '-'
	mov byte ptr [edi], al          ;put the first char in the dest str
	inc edi
	
	;---------------------------
	;get the exponent of the dbl
	;---------------------------
	fld qword ptr [dbl]             ;load the possitive dbl
	fldlg2                          ;load log10(2)
	fxch                            ;set up registers for next operation
	fyl2x                           ;st = log10(2)*log2(dbl) == log10(dbl)
	 
	fstcw word ptr [oldCW]          ;store the crnt control word
	mov ax, word ptr [oldCW]   
	or ah, 00000100b                ;set the new rounding mode, set RC field to "round down" (10b)
	mov word ptr [newCW], ax        ;store the modyfied CW and...
	fldcw word ptr [newCW]          ;...load it
	frndint                         ;now round the st to get the exponent
	fldcw word ptr [oldCW]          ;restore the prev CW
	 
	fistp dword ptr [exponent]      ;load the exponent
	
	;---------------------------------------------------------
	;get the decimal. 1) dbl=dbl^10^(-exp) (px. 12.3->.123)...
	;...2) dbl=dbl*10^7 (.123->1230000.) 3) decimal=rnd(dbl)
	;---------------------------------------------------------
	fild  dword ptr [exponent]      ;push the exponent to the regstack
	fchs                            ;exponent = -exponent
	
	fldl2t                          ;st = log2(10) * exponent
	fmulp st(1), st            
      
	fld st(0)                       ;at(1) = st
	frndint                    
	fsub st(1), st                  ;st = (log2(10)*exponent) - rnd(log2(10)*exponent)
	fxch

	f2xm1                                             
	fld1
	faddp st(1), st                 ;st = 2^((log2(10)*exponent) - rnd(log2(10)*exponent))

	fscale                          ;st = 10^exponent
	
	fstp st(1)                      ;push reg stack
	fld qword ptr [dbl]
	fmulp st(1), st                 ;st = 10^exponent (if dbl==0.001 then st==0.1)
	 
	fmul dword ptr [f10pwr7]        ;st = (10^exponent)*(10^7) (if dbl==0.0023 then st==2300000.0)
	fistp dword ptr [decimal]
	
	;---------------------------------------------------
	;using the decimal and the exponent build the string
	;---------------------------------------------------
	inc edi                         ;leave a blank byte
	invoke gUIntToStr, edi, decimal ;and then call gUIntToStr. ex. "+?1234..."
	mov cl, byte ptr [edi]          ;now move the first byte to the black byte. "+11234..."
	mov byte ptr [edi-1], cl   
	mov byte ptr [edi], '.'         ;put the '.' and the normalized mantissa is in the string!! "+1.234..."
	add edi, eax                    ;remember that the gUIntToStr returns the size of the source string
	
	mov byte ptr [edi], 'e'
	inc edi                         ;put the 'e'
	
	cmp exponent, 0                 ;if exponent>0 then put the '+' in the pcDest (without brunch)
	jl @F
		mov byte ptr [edi], '+'
		inc edi
	@@:
	
	invoke gIntToStr, edi, exponent ;put the exponent
	
	;------------------------------
	;compute the return value (eax)
	;------------------------------
	add eax, edi                    ;eax has allready the return val from gIntToStr
	sub eax, dword ptr [pcDest]     ;sub the begining of the string
	
	return:
	pop ecx
	pop esi
	pop edi
	ret
gDoubleToStr endp

end